/*
 * 	drivers/char/filtror.c
 *
 *
 *	FiltrorCon v0.0.1:	A Filtror console device driver
 *
 *	(c) Copyright 2002	Milosch Meriac <xboxlinux@meriac.de>
 *				All Rights Reserved.
 *
 *				http://xboxlinux.meriac.de
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 *	Neither Milosch Meriac nor Bitmanufaktur admit liability nor provide 
 *	warranty for any of this software. This material is provided 
 *	"AS-IS" and at no charge.	
 *
 *	(c) Copyright 2002	Milosch Meriac <xboxlinux@meriac.de>
 *
 *	This is a filtror-device console driver for linux. It has devfs-
 *	capabilities and provides a device called "/dev/filtror".
 *
 *  20020813 Milosch Meriac <xboxlinux@meriac.de>
 *	Goin' public. v0.0.1 Tested with busybox (non-module).
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/io.h>

#define FILTROR_MINOR 64+16
#define MAX_KB_RETRIES 10000
#define MAX_BUFFER_POS 0x7FB
#define FLAG_BUFFER_READY 0x8000

//#pragma pack(1)
typedef struct
{
    char buffer[MAX_BUFFER_POS+1];
    u_short checksum;
    u_short size;
} TSharedFiltrorTransferBuffer;

typedef struct
{
    TSharedFiltrorTransferBuffer server,client;
} TSharedFiltrorMemory;
//#pragma pack(8)

static volatile TSharedFiltrorMemory* shared = (TSharedFiltrorMemory*)0xFFFFE000;
static devfs_handle_t hDevFs = NULL;
const char welcome_message[]=KERN_INFO "Filtror console driver v0.1b activated.\n";

static int nCountFiltrorConsoles=0;
static int nOutputCount=0;

static unsigned char baInputBuffer[2048];
static int nFiltrorInputBufferPos=0;
static int nFiltrorInputBufferMax=0;

/**************************************************************************
 *
 *  Filtror lowlevel interface functions
 *
 **************************************************************************/
 
void set_cpu_cache(int active)
{
    active=(active & 1) << 30;
    __asm__ (
	"pushf				;"
	"cli				;"
	"mov	%%cr0, %%eax		;"
	"and	$0xBFFFFFFF, %%eax	;"
	"or	%%ebx, %%eax		;"
	"mov	%%eax, %%cr0		;"
	"wbinvd				;" 
	"popf				"
	: : "ebx" (active) : "%eax"
    );
}

inline u_short filtror_checksum(const char *s,unsigned count)
{
    u_short sum=0;
    
    for(;count--;s++)
	sum+=*s;
    return sum;
}

static int filtror_lowlevel_open(void)
{    
    int res = -1;
    const u_short TestPattern = 0x4223;
    const unsigned count = sizeof(welcome_message)-1;

    set_cpu_cache(0);

				    memset((void*)shared,0,sizeof(*shared));
    if(!(shared->client.size || shared->server.size))
    {
	shared->server.size = shared->client.size = TestPattern;
	if((shared->server.size==TestPattern)&&(shared->client.size==TestPattern))
	{

			shared->client.size = shared->server.size = 0;

	    //
	    // Static copy of welcome message to filtror
	    //
	    memcpy((void*)shared->server.buffer,(void*)&welcome_message,count);
	    shared->server.checksum = filtror_checksum(welcome_message,count);
	    shared->server.size = count|FLAG_BUFFER_READY;
	    res=0;
	}

		}
    set_cpu_cache(1);

    if(res)
        printk(KERN_ERR "  Filtror: Can't find Filtror device.\n");
    else
        printk(KERN_DEBUG "  Filtror: Init Filtror device.\n");

    return res;
}

static void filtror_lowlevel_release(void)
{
    printk(KERN_DEBUG "  Filtror: Close Filtror device.\n");
}

static ssize_t filtror_lowlevel_write(const char *s, unsigned count)
{
  //  int pos=0,end,ready,flags,remain;

	int nPos=0, nUse;
    
		set_cpu_cache(0);

	while(nPos<count) {

		while(shared->server.size & FLAG_BUFFER_READY) ;  // spin waiting for last data to get eaten by PC

		nUse=(count-nPos);
		if(nUse>MAX_BUFFER_POS) nUse=MAX_BUFFER_POS;

		memcpy((void*)(&shared->server.buffer[0]),s+nPos, nUse);
		shared->server.checksum=filtror_checksum(s+nPos, nUse);
	 shared->server.size=nUse|FLAG_BUFFER_READY;

		nPos+=nUse;
	}

/*
    flags=shared->server.size;
    ready=(flags&FLAG_BUFFER_READY)>0;
    pos=flags & !FLAG_BUFFER_READY;

    end=pos+count;
    if(end>MAX_BUFFER_POS)
    {
	remain=end-(MAX_BUFFER_POS+1);
	count-=remain;
	end=MAX_BUFFER_POS;
    }
    else
	remain=0;

    if(count)
    {
	if(!ready && pos)
	{
	    while(!(shared->server.size & FLAG_BUFFER_READY));
	    ready=1;
	    pos=0;
	}

	if(ready || !flags)
	{
//	    shared->server.size=FLAG_BUFFER_READY;
	    memcpy((void*)(&shared->server.buffer[0]+pos),s,count);
	    if(pos)
		shared->server.checksum=filtror_checksum((char*)&shared->server.buffer,end);
	    else
		shared->server.checksum=filtror_checksum(s,count);
	    shared->server.size=end|FLAG_BUFFER_READY;
	}
    }
 */

    set_cpu_cache(1);

    return count;
}

static ssize_t filtror_lowlevel_read(char *s, unsigned countMax)
{
	int nCountFromBuffer=0, nCountFromBufferFresh=0;
	unsigned int n;

		// first see if we can fulfil the request from the buffer, partially or wholly

	if(nFiltrorInputBufferPos<nFiltrorInputBufferMax) { // something from the buffer
		nCountFromBuffer=nFiltrorInputBufferMax-nFiltrorInputBufferPos;
		if(nCountFromBuffer > countMax) nCountFromBuffer=countMax;
		memcpy(s, &baInputBuffer[0], nCountFromBuffer);
		nFiltrorInputBufferPos+=nCountFromBuffer;
		if(nCountFromBuffer == countMax) return countMax;  // wholly satisfied by the buffer
	}

			// we only get here if the buffer is empty, in fact - we may have just used it up and want more

	n = shared->client.size;

	if(n&FLAG_BUFFER_READY) {
		nFiltrorInputBufferMax = (n&0x7ffff);
		if(nFiltrorInputBufferMax > sizeof(baInputBuffer)) nFiltrorInputBufferMax=sizeof(baInputBuffer);
		memcpy(&baInputBuffer[0], (void*)&shared->client.buffer, nFiltrorInputBufferMax);
		shared->client.size=0; // mark as read
		nFiltrorInputBufferPos=0;
		if(nFiltrorInputBufferMax < (countMax- nCountFromBuffer )) {
			nCountFromBufferFresh = nFiltrorInputBufferMax;
		} else {
			nCountFromBufferFresh = (countMax-nCountFromBuffer);
		}
		memcpy(s+nCountFromBuffer, &baInputBuffer[0], nCountFromBufferFresh);
		nCountFromBuffer+=nCountFromBufferFresh;
		nFiltrorInputBufferPos+=nCountFromBufferFresh;
	}

	return nCountFromBuffer;

    /*
		int pos,ready,flags;

    if(count)
    {
	set_cpu_cache(0);

	flags=shared->client.size;
	ready=(flags&FLAG_BUFFER_READY)>0;
	pos=flags & !FLAG_BUFFER_READY;
	if(ready && pos)
	{
	    if(count>pos)
		count=pos;
	    memcpy((void*)s,(void*)&shared->client.buffer,count);
	    shared->client.size=0;
	}

	set_cpu_cache(1);
    }

    return count;
*/
}

/**************************************************************************
 *
 *  Console support functions
 *
 **************************************************************************/

static void filtror_console_write(struct console *co, const char *s, unsigned count)
{
    filtror_lowlevel_write(s,count);
}

static kdev_t filtror_console_device(struct console *c)
{
    return MKDEV(MISC_MAJOR,FILTROR_MINOR + c->index);
}

static int __init filtror_console_setup(struct console *co, char *options)
{
    printk(KERN_DEBUG "  Filtror: console_setup.\n");

    if(co->index)
        return -1;
	
    co->cflag=CREAD|HUPCL|CLOCAL|B9600|CS8;
        
    return 0;
}

static struct console filtror_cons = {
    name:	"filtror",
    write:	filtror_console_write,
    device:	filtror_console_device,
    setup:	filtror_console_setup,
    flags:	CON_PRINTBUFFER,
    index:	-1,
};

/**************************************************************************
 *
 *  /dev/filtror - Device support functions
 *
 **************************************************************************/

static ssize_t filtror_fread(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
    return filtror_lowlevel_read(buffer,count);
}

static ssize_t filtror_fwrite(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
    return filtror_lowlevel_write(buffer,count);
}

static int filtror_fopen(struct inode *inode, struct file *file)
{
    MOD_INC_USE_COUNT;

    return 0;
}

static int filtror_frelease(struct inode *inode, struct file *file)
{
    MOD_DEC_USE_COUNT;    
    
    return 0;
}

static struct file_operations filtror_fops = {
    owner:	THIS_MODULE,
    read:	filtror_fread,
    write:	filtror_fwrite,
    open:	filtror_fopen,
    release:	filtror_frelease,
};
    
static struct miscdevice filtror_dev = {
    minor:	FILTROR_MINOR,
    name:	"filtror",
    fops:	&filtror_fops,
};   
    
/**************************************************************************
 *
 *  Module support functions
 *
 **************************************************************************/

static int __init filtror_init_module(void)
{
    int ret;

//		nCountFiltrorConsoles++;
//		if(nCountFiltrorConsoles>1) return 0;

		shared = (volatile TSharedFiltrorMemory*)ioremap(0xff0fe000,0x1000);

    printk(KERN_DEBUG "  Filtror: init_module, base is at %x\n", shared);

    if(filtror_lowlevel_open())
    {
        printk(KERN_ERR "Can't find Filtror device.\n");
	return -ENODEV;
    }

    printk(KERN_DEBUG "  Filtror: managed to open device\n");

    ret = misc_register(&filtror_dev);
    if (ret)
    {
        printk(KERN_ERR "Filtror device can't be registered.\n");
	return ret;
    }

    printk(KERN_DEBUG "  Filtror: misc_register is okay\n");

    hDevFs = devfs_register(
	NULL,
	filtror_dev.name,
	DEVFS_FL_NONE,
	MISC_MAJOR,
	FILTROR_MINOR,
	S_IRUGO | S_IWUGO | S_IFCHR,
	&filtror_fops,
	NULL
	);

    printk(welcome_message);
    printk(KERN_DEBUG "  Filtror: init_module: register_console.\n");
    register_console(&filtror_cons);
    printk(KERN_DEBUG "  Filtror: register_console done.\n");
    return 0;
}

static void filtror_cleanup_module(void)
{
    printk(KERN_DEBUG "  Filtror: cleanup_module\n");

    if(hDevFs)
        devfs_unregister(hDevFs);

    unregister_console(&filtror_cons);

    filtror_lowlevel_release();

    printk(KERN_DEBUG "  Filtror: cleanup_module done.\n");
}

module_init(filtror_init_module);
module_exit(filtror_cleanup_module);

MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
