#ifndef _LINUX_FATX_FS_H
#define _LINUX_FATX_FS_H

#ifdef __KERNEL__

/*
 * The MS-DOS filesystem constants/structures
 */
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/fd.h>
#include <linux/dcache.h>

#include <asm/byteorder.h>

#include <linux/msdos_fs.h>

#define FATX_BOOTBLOCK_MAGIC cpu_to_le32(0x58544146)
#define FATX_MAX_NAME_LENGTH 42
#define FATX_DIR_BITS 6

//FF == end of directory info, E5 == deleted entry
#define FATX_IS_FREE(de) ((de)->name_length==DELETED_FLAG)
#define FATX_END_OF_DIR(de) ((de)->name_length==0xFF)

struct fatx_boot_sector {
        __u32	magic;		/* "FATX" */
	__u32	volume_id;	/* Volume ID */
        __u32	cluster_size;	/* sectors/cluster */
	__u16	fats;		/* number of FATs */
	__u32	unknown;
};

struct fatx_dir_entry {
        __u8	name_length;	/* length of filename (bytes) */
	__u8	attr;		/* attribute bits */
        __s8	name[42];	/* filename */
	__u32	start;		/* time, date and first cluster */
	__u32	size;		/* file size (in bytes) */
	__u16	time,date;	/* time, date */
	__u16	ctime,cdate;	/* Creation time */
	__u16	atime,adate;	/* Last access time */
};

/* fatx/namei.c - these are for the xbox's FATX */
extern struct dentry *fatx_lookup(struct inode *dir, struct dentry *);
extern int fatx_create(struct inode *dir, struct dentry *dentry, int mode);
extern int fatx_rmdir(struct inode *dir, struct dentry *dentry);
extern int fatx_unlink(struct inode *dir, struct dentry *dentry);
extern int fatx_mkdir(struct inode *dir, struct dentry *dentry, int mode);
extern int fatx_rename(struct inode *old_dir, struct dentry *old_dentry,
		       struct inode *new_dir, struct dentry *new_dentry);
extern struct super_block *fatx_read_super(struct super_block *sb, void *data,
					   int silent);

/* fatx/fatxfs_syms.c */
extern unsigned int fatx_debug;

extern struct file_system_type fatx_fs_type;

extern int fatx__get_entry(
		struct inode *dir, 
		loff_t *pos,
		struct buffer_head **bh,
		struct fatx_dir_entry **de,
		int *ino );

static __inline__ int fatx_get_entry(
		struct inode *dir,
		loff_t *pos,
		struct buffer_head **bh,
		struct fatx_dir_entry **de,
		int *ino )
{
	
	/* Fast stuff first */
/*	if (*bh && *de &&
	    ((*de - (struct fatx_dir_entry *)(*bh)->b_data)>>MSDOS_SB(dir->i_sb)->dir_per_block_bits) < MSDOS_SB(dir->i_sb)->dir_per_block - 1) {
		*pos += sizeof(struct fatx_dir_entry);
		(*de)++;
		(*ino)++;
		return 0;
	}
*/	return fatx__get_entry(dir,pos,bh,de,ino);
}

extern int fatx_scan(
		struct inode *dir, 
		const char *name,
		int name_length,
		struct buffer_head **res_bh,
		struct fatx_dir_entry **res_de,
		int *ino);

extern struct inode *fatx_build_inode(
		struct super_block *sb,
		struct fatx_dir_entry *de,
		int ino,
		int *res );
		
/* miscelaneous support code for fatx fs */
extern int fatx_date_dos2unix( unsigned short, unsigned short );
extern void fatx_date_unix2dos( int, unsigned short *, unsigned short * );

static inline unsigned char
fatx_tolower(struct nls_table *t, unsigned char c)
{
	unsigned char nc = t->charset2lower[c];

	return nc ? nc : c;
}

static inline unsigned char
fatx_toupper(struct nls_table *t, unsigned char c)
{
	unsigned char nc = t->charset2upper[c];

	return nc ? nc : c;
}

static inline int
fatx_strnicmp(	struct nls_table *t,
		const unsigned char *s1,
		const unsigned char *s2,
		int len )
{
	while(len--) {
		if (fatx_tolower(t, *s1++) != fatx_tolower(t, *s2++))
			return 1;
	}
	return 0;
}

/* directory code for fatx fs */
extern int fatx_do_add_entry(
		struct inode *dir,
		struct buffer_head **bh,
		struct fatx_dir_entry **de,
		int *ino);
		
extern int fatx_new_dir(struct inode *dir, struct inode *parent);
extern int fatx_dir_empty(struct inode *dir);
extern int fatx_subdirs(struct inode *dir);
		
extern struct file_operations fatx_dir_operations;
extern struct dentry_operations fatx_dentry_operations;
#endif /* __KERNEL__ */

#endif /* _LINUX_FATX_FS_H */
